@@ -4,8 +4,8 @@ extern crate core;
 
 use std::sync::OnceLock;
 
+use jiter::StringCacheMode;
 use pyo3::exceptions::PyTypeError;
-use pyo3::types::{PyByteArray, PyBytes, PyString};
 use pyo3::{prelude::*, sync::GILOnceCell};
 
 // parse this first to get access to the contained macro
@@ -37,17 +37,33 @@ pub use serializers::{
 };
 pub use validators::{validate_core_schema, PySome, SchemaValidator};
 
-#[pyfunction(signature = (data, *, allow_inf_nan=true))]
-pub fn from_json(py: Python, data: &PyAny, allow_inf_nan: bool) -> PyResult<PyObject> {
-    if let Ok(py_bytes) = data.downcast::<PyBytes>() {
-        jiter::python_parse(py, py_bytes.as_bytes(), allow_inf_nan)
-    } else if let Ok(py_str) = data.downcast::<PyString>() {
-        jiter::python_parse(py, py_str.to_str()?.as_bytes(), allow_inf_nan)
-    } else if let Ok(py_byte_array) = data.downcast::<PyByteArray>() {
-        jiter::python_parse(py, &py_byte_array.to_vec(), allow_inf_nan)
-    } else {
-        Err(PyTypeError::new_err("Expected bytes, bytearray or str"))
-    }
+use crate::input::Input;
+
+#[derive(FromPyObject)]
+pub enum CacheStringsArg {
+    Bool(bool),
+    Literal(StringCacheMode),
+}
+
+#[pyfunction(signature = (data, *, allow_inf_nan=true, cache_strings=CacheStringsArg::Bool(true), allow_partial=false))]
+pub fn from_json<'py>(
+    py: Python<'py>,
+    data: &Bound<'_, PyAny>,
+    allow_inf_nan: bool,
+    cache_strings: CacheStringsArg,
+    allow_partial: bool,
+) -> PyResult<Bound<'py, PyAny>> {
+    let v_match = data
+        .validate_bytes(false)
+        .map_err(|_| PyTypeError::new_err("Expected bytes, bytearray or str"))?;
+    let json_either_bytes = v_match.into_inner();
+    let json_bytes = json_either_bytes.as_slice();
+    let cache_mode = match cache_strings {
+        CacheStringsArg::Bool(b) => b.into(),
+        CacheStringsArg::Literal(mode) => mode,
+    };
+    jiter::python_parse(py, json_bytes, allow_inf_nan, cache_mode, allow_partial)
+        .map_err(|e| jiter::map_json_error(json_bytes, &e))
 }
 
 pub fn get_pydantic_core_version() -> &'static str {
@@ -70,7 +86,7 @@ fn get_pydantic_version(py: Python<'_>) -> Option<&'static str> {
 
     PYDANTIC_VERSION
         .get_or_init(py, || {
-            py.import("pydantic")
+            py.import_bound("pydantic")
                 .and_then(|pydantic| pydantic.getattr("__version__")?.extract())
                 .ok()
         })
@@ -86,7 +102,7 @@ pub fn build_info() -> String {
 }
 
 #[pymodule]
-fn _pydantic_core(py: Python, m: &PyModule) -> PyResult<()> {
+fn _pydantic_core(py: Python, m: &Bound<'_, PyModule>) -> PyResult<()> {
     m.add("__version__", get_pydantic_core_version())?;
     m.add("build_profile", env!("PROFILE"))?;
     m.add("build_info", build_info())?;
