@@ -10,8 +10,8 @@ use crate::py_gc::PyGcTraverse;
 
 use config::SerializationConfig;
 pub use errors::{PydanticSerializationError, PydanticSerializationUnexpectedValue};
-use extra::{CollectWarnings, SerRecursionGuard};
-pub(crate) use extra::{Extra, SerMode, SerializationState};
+use extra::{CollectWarnings, SerRecursionState};
+pub(crate) use extra::{DuckTypingSerMode, Extra, SerMode, SerializationState};
 pub use shared::CombinedSerializer;
 use shared::{to_json_bytes, BuildSerializer, TypeSerializer};
 
@@ -52,9 +52,11 @@ impl SchemaSerializer {
         exclude_defaults: bool,
         exclude_none: bool,
         round_trip: bool,
-        rec_guard: &'a SerRecursionGuard,
+        rec_guard: &'a SerRecursionState,
         serialize_unknown: bool,
-        fallback: Option<&'a PyAny>,
+        fallback: Option<&'a Bound<'a, PyAny>>,
+        duck_typing_ser_mode: DuckTypingSerMode,
+        context: Option<&'a Bound<'a, PyAny>>,
     ) -> Extra<'b> {
         Extra::new(
             py,
@@ -69,6 +71,8 @@ impl SchemaSerializer {
             rec_guard,
             serialize_unknown,
             fallback,
+            duck_typing_ser_mode,
+            context,
         )
     }
 }
@@ -76,7 +80,7 @@ impl SchemaSerializer {
 #[pymethods]
 impl SchemaSerializer {
     #[new]
-    pub fn py_new(py: Python, schema: &PyDict, config: Option<&PyDict>) -> PyResult<Self> {
+    pub fn py_new(schema: Bound<'_, PyDict>, config: Option<&Bound<'_, PyDict>>) -> PyResult<Self> {
         let mut definitions_builder = DefinitionsBuilder::new();
         let serializer = CombinedSerializer::build(schema.downcast()?, config, &mut definitions_builder)?;
         Ok(Self {
@@ -84,9 +88,9 @@ impl SchemaSerializer {
             definitions: definitions_builder.finish()?,
             expected_json_size: AtomicUsize::new(1024),
             config: SerializationConfig::from_config(config)?,
-            py_schema: schema.into_py(py),
+            py_schema: schema.into(),
             py_config: match config {
-                Some(c) if !c.is_empty() => Some(c.into_py(py)),
+                Some(c) if !c.is_empty() => Some(c.clone().into()),
                 _ => None,
             },
         })
@@ -95,25 +99,28 @@ impl SchemaSerializer {
     #[allow(clippy::too_many_arguments)]
     #[pyo3(signature = (value, *, mode = None, include = None, exclude = None, by_alias = true,
         exclude_unset = false, exclude_defaults = false, exclude_none = false, round_trip = false, warnings = true,
-        fallback = None))]
+        fallback = None, serialize_as_any = false, context = None))]
     pub fn to_python(
         &self,
         py: Python,
-        value: &PyAny,
+        value: &Bound<'_, PyAny>,
         mode: Option<&str>,
-        include: Option<&PyAny>,
-        exclude: Option<&PyAny>,
+        include: Option<&Bound<'_, PyAny>>,
+        exclude: Option<&Bound<'_, PyAny>>,
         by_alias: bool,
         exclude_unset: bool,
         exclude_defaults: bool,
         exclude_none: bool,
         round_trip: bool,
         warnings: bool,
-        fallback: Option<&PyAny>,
+        fallback: Option<&Bound<'_, PyAny>>,
+        serialize_as_any: bool,
+        context: Option<&Bound<'_, PyAny>>,
     ) -> PyResult<PyObject> {
         let mode: SerMode = mode.into();
         let warnings = CollectWarnings::new(warnings);
-        let rec_guard = SerRecursionGuard::default();
+        let rec_guard = SerRecursionState::default();
+        let duck_typing_ser_mode = DuckTypingSerMode::from_bool(serialize_as_any);
         let extra = self.build_extra(
             py,
             &mode,
@@ -126,6 +133,8 @@ impl SchemaSerializer {
             &rec_guard,
             false,
             fallback,
+            duck_typing_ser_mode,
+            context,
         );
         let v = self.serializer.to_python(value, include, exclude, &extra)?;
         warnings.final_check(py)?;
@@ -135,24 +144,27 @@ impl SchemaSerializer {
     #[allow(clippy::too_many_arguments)]
     #[pyo3(signature = (value, *, indent = None, include = None, exclude = None, by_alias = true,
         exclude_unset = false, exclude_defaults = false, exclude_none = false, round_trip = false, warnings = true,
-        fallback = None))]
+        fallback = None, serialize_as_any = false, context = None))]
     pub fn to_json(
         &self,
         py: Python,
-        value: &PyAny,
+        value: &Bound<'_, PyAny>,
         indent: Option<usize>,
-        include: Option<&PyAny>,
-        exclude: Option<&PyAny>,
+        include: Option<&Bound<'_, PyAny>>,
+        exclude: Option<&Bound<'_, PyAny>>,
         by_alias: bool,
         exclude_unset: bool,
         exclude_defaults: bool,
         exclude_none: bool,
         round_trip: bool,
         warnings: bool,
-        fallback: Option<&PyAny>,
+        fallback: Option<&Bound<'_, PyAny>>,
+        serialize_as_any: bool,
+        context: Option<&Bound<'_, PyAny>>,
     ) -> PyResult<PyObject> {
         let warnings = CollectWarnings::new(warnings);
-        let rec_guard = SerRecursionGuard::default();
+        let rec_guard = SerRecursionState::default();
+        let duck_typing_ser_mode = DuckTypingSerMode::from_bool(serialize_as_any);
         let extra = self.build_extra(
             py,
             &SerMode::Json,
@@ -165,6 +177,8 @@ impl SchemaSerializer {
             &rec_guard,
             false,
             fallback,
+            duck_typing_ser_mode,
+            context,
         );
         let bytes = to_json_bytes(
             value,
@@ -179,11 +193,11 @@ impl SchemaSerializer {
         warnings.final_check(py)?;
 
         self.expected_json_size.store(bytes.len(), Ordering::Relaxed);
-        let py_bytes = PyBytes::new(py, &bytes);
+        let py_bytes = PyBytes::new_bound(py, &bytes);
         Ok(py_bytes.into())
     }
 
-    pub fn __reduce__(slf: &PyCell<Self>) -> PyResult<(PyObject, (PyObject, PyObject))> {
+    pub fn __reduce__(slf: &Bound<Self>) -> PyResult<(PyObject, (PyObject, PyObject))> {
         // Enables support for `pickle` serialization.
         let py = slf.py();
         let cls = slf.get_type().into();
@@ -213,22 +227,27 @@ impl SchemaSerializer {
 #[pyfunction]
 #[pyo3(signature = (value, *, indent = None, include = None, exclude = None, by_alias = true,
     exclude_none = false, round_trip = false, timedelta_mode = "iso8601", bytes_mode = "utf8",
-    serialize_unknown = false, fallback = None))]
+    inf_nan_mode = "constants", serialize_unknown = false, fallback = None, serialize_as_any = false,
+    context = None))]
 pub fn to_json(
     py: Python,
-    value: &PyAny,
+    value: &Bound<'_, PyAny>,
     indent: Option<usize>,
-    include: Option<&PyAny>,
-    exclude: Option<&PyAny>,
+    include: Option<&Bound<'_, PyAny>>,
+    exclude: Option<&Bound<'_, PyAny>>,
     by_alias: bool,
     exclude_none: bool,
     round_trip: bool,
     timedelta_mode: &str,
     bytes_mode: &str,
+    inf_nan_mode: &str,
     serialize_unknown: bool,
-    fallback: Option<&PyAny>,
+    fallback: Option<&Bound<'_, PyAny>>,
+    serialize_as_any: bool,
+    context: Option<&Bound<'_, PyAny>>,
 ) -> PyResult<PyObject> {
-    let state = SerializationState::new(timedelta_mode, bytes_mode)?;
+    let state = SerializationState::new(timedelta_mode, bytes_mode, inf_nan_mode)?;
+    let duck_typing_ser_mode = DuckTypingSerMode::from_bool(serialize_as_any);
     let extra = state.extra(
         py,
         &SerMode::Json,
@@ -237,32 +256,39 @@ pub fn to_json(
         round_trip,
         serialize_unknown,
         fallback,
+        duck_typing_ser_mode,
+        context,
     );
     let serializer = type_serializers::any::AnySerializer.into();
     let bytes = to_json_bytes(value, &serializer, include, exclude, &extra, indent, 1024)?;
     state.final_check(py)?;
-    let py_bytes = PyBytes::new(py, &bytes);
+    let py_bytes = PyBytes::new_bound(py, &bytes);
     Ok(py_bytes.into())
 }
 
 #[allow(clippy::too_many_arguments)]
 #[pyfunction]
 #[pyo3(signature = (value, *, include = None, exclude = None, by_alias = true, exclude_none = false, round_trip = false,
-    timedelta_mode = "iso8601", bytes_mode = "utf8", serialize_unknown = false, fallback = None))]
+    timedelta_mode = "iso8601", bytes_mode = "utf8", inf_nan_mode = "constants", serialize_unknown = false, fallback = None,
+    serialize_as_any = false, context = None))]
 pub fn to_jsonable_python(
     py: Python,
-    value: &PyAny,
-    include: Option<&PyAny>,
-    exclude: Option<&PyAny>,
+    value: &Bound<'_, PyAny>,
+    include: Option<&Bound<'_, PyAny>>,
+    exclude: Option<&Bound<'_, PyAny>>,
     by_alias: bool,
     exclude_none: bool,
     round_trip: bool,
     timedelta_mode: &str,
     bytes_mode: &str,
+    inf_nan_mode: &str,
     serialize_unknown: bool,
-    fallback: Option<&PyAny>,
+    fallback: Option<&Bound<'_, PyAny>>,
+    serialize_as_any: bool,
+    context: Option<&Bound<'_, PyAny>>,
 ) -> PyResult<PyObject> {
-    let state = SerializationState::new(timedelta_mode, bytes_mode)?;
+    let state = SerializationState::new(timedelta_mode, bytes_mode, inf_nan_mode)?;
+    let duck_typing_ser_mode = DuckTypingSerMode::from_bool(serialize_as_any);
     let extra = state.extra(
         py,
         &SerMode::Json,
@@ -271,6 +297,8 @@ pub fn to_jsonable_python(
         round_trip,
         serialize_unknown,
         fallback,
+        duck_typing_ser_mode,
+        context,
     );
     let v = infer::infer_to_python(value, include, exclude, &extra)?;
     state.final_check(py)?;
