@@ -1,5 +1,6 @@
 use pyo3::intern;
 use pyo3::prelude::*;
+use pyo3::types::PyString;
 use pyo3::types::{PyDict, PyList};
 
 use crate::definitions::DefinitionRef;
@@ -18,24 +19,24 @@ impl BuildValidator for DefinitionsValidatorBuilder {
     const EXPECTED_TYPE: &'static str = "definitions";
 
     fn build(
-        schema: &PyDict,
-        config: Option<&PyDict>,
+        schema: &Bound<'_, PyDict>,
+        config: Option<&Bound<'_, PyDict>>,
         definitions: &mut DefinitionsBuilder<CombinedValidator>,
     ) -> PyResult<CombinedValidator> {
         let py = schema.py();
 
-        let schema_definitions: &PyList = schema.get_as_req(intern!(py, "definitions"))?;
+        let schema_definitions: Bound<'_, PyList> = schema.get_as_req(intern!(py, "definitions"))?;
 
         for schema_definition in schema_definitions {
             let reference = schema_definition
-                .extract::<&PyDict>()?
+                .extract::<Bound<'_, PyDict>>()?
                 .get_as_req::<String>(intern!(py, "ref"))?;
-            let validator = build_validator(schema_definition, config, definitions)?;
+            let validator = build_validator(&schema_definition, config, definitions)?;
             definitions.add_definition(reference, validator)?;
         }
 
-        let inner_schema: &PyAny = schema.get_as_req(intern!(py, "schema"))?;
-        build_validator(inner_schema, config, definitions)
+        let inner_schema = schema.get_as_req(intern!(py, "schema"))?;
+        build_validator(&inner_schema, config, definitions)
     }
 }
 
@@ -54,13 +55,13 @@ impl BuildValidator for DefinitionRefValidator {
     const EXPECTED_TYPE: &'static str = "definition-ref";
 
     fn build(
-        schema: &PyDict,
-        _config: Option<&PyDict>,
+        schema: &Bound<'_, PyDict>,
+        _config: Option<&Bound<'_, PyDict>>,
         definitions: &mut DefinitionsBuilder<CombinedValidator>,
     ) -> PyResult<CombinedValidator> {
-        let schema_ref = schema.get_as_req(intern!(schema.py(), "schema_ref"))?;
+        let schema_ref: Bound<'_, PyString> = schema.get_as_req(intern!(schema.py(), "schema_ref"))?;
 
-        let definition = definitions.get_definition(schema_ref);
+        let definition = definitions.get_definition(schema_ref.to_str()?);
         Ok(Self::new(definition).into())
     }
 }
@@ -68,15 +69,16 @@ impl BuildValidator for DefinitionRefValidator {
 impl_py_gc_traverse!(DefinitionRefValidator {});
 
 impl Validator for DefinitionRefValidator {
-    fn validate<'data>(
+    fn validate<'py>(
         &self,
-        py: Python<'data>,
-        input: &'data impl Input<'data>,
-        state: &mut ValidationState,
+        py: Python<'py>,
+        input: &(impl Input<'py> + ?Sized),
+        state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<PyObject> {
         self.definition.read(|validator| {
             let validator = validator.unwrap();
-            if let Some(id) = input.identity() {
+            if let Some(id) = input.as_python().map(py_identity) {
+                // Python objects can be cyclic, so need recursion guard
                 let Ok(mut guard) = RecursionGuard::new(state, id, self.definition.id()) else {
                     return Err(ValError::new(ErrorTypeDefaults::RecursionLoop, input));
                 };
@@ -87,24 +89,20 @@ impl Validator for DefinitionRefValidator {
         })
     }
 
-    fn validate_assignment<'data>(
+    fn validate_assignment<'py>(
         &self,
-        py: Python<'data>,
-        obj: &'data PyAny,
-        field_name: &'data str,
-        field_value: &'data PyAny,
-        state: &mut ValidationState,
+        py: Python<'py>,
+        obj: &Bound<'py, PyAny>,
+        field_name: &str,
+        field_value: &Bound<'py, PyAny>,
+        state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<PyObject> {
         self.definition.read(|validator| {
             let validator = validator.unwrap();
-            if let Some(id) = obj.identity() {
-                let Ok(mut guard) = RecursionGuard::new(state, id, self.definition.id()) else {
-                    return Err(ValError::new(ErrorTypeDefaults::RecursionLoop, obj));
-                };
-                validator.validate_assignment(py, obj, field_name, field_value, guard.state())
-            } else {
-                validator.validate_assignment(py, obj, field_name, field_value, state)
-            }
+            let Ok(mut guard) = RecursionGuard::new(state, py_identity(obj), self.definition.id()) else {
+                return Err(ValError::new(ErrorTypeDefaults::RecursionLoop, obj));
+            };
+            validator.validate_assignment(py, obj, field_name, field_value, guard.state())
         })
     }
 
@@ -112,3 +110,7 @@ impl Validator for DefinitionRefValidator {
         self.definition.get_or_init_name(|v| v.get_name().into())
     }
 }
+
+fn py_identity(obj: &Bound<'_, PyAny>) -> usize {
+    obj.as_ptr() as usize
+}
