@@ -3,7 +3,7 @@ use pyo3::prelude::*;
 use pyo3::types::{PyDict, PyString};
 use regex::Regex;
 
-use crate::build_tools::{is_strict, py_schema_error_type, schema_or_config};
+use crate::build_tools::{is_strict, py_schema_error_type, schema_or_config, schema_or_config_same};
 use crate::errors::{ErrorType, ValError, ValResult};
 use crate::input::Input;
 use crate::tools::SchemaDict;
@@ -20,8 +20,8 @@ impl BuildValidator for StrValidator {
     const EXPECTED_TYPE: &'static str = "str";
 
     fn build(
-        schema: &PyDict,
-        config: Option<&PyDict>,
+        schema: &Bound<'_, PyDict>,
+        config: Option<&Bound<'_, PyDict>>,
         _definitions: &mut DefinitionsBuilder<CombinedValidator>,
     ) -> PyResult<CombinedValidator> {
         let con_str_validator = StrConstrainedValidator::build(schema, config)?;
@@ -41,15 +41,15 @@ impl BuildValidator for StrValidator {
 impl_py_gc_traverse!(StrValidator {});
 
 impl Validator for StrValidator {
-    fn validate<'data>(
+    fn validate<'py>(
         &self,
-        py: Python<'data>,
-        input: &'data impl Input<'data>,
-        state: &mut ValidationState,
-    ) -> ValResult<'data, PyObject> {
+        py: Python<'py>,
+        input: &(impl Input<'py> + ?Sized),
+        state: &mut ValidationState<'_, 'py>,
+    ) -> ValResult<PyObject> {
         input
             .validate_str(state.strict_or(self.strict), self.coerce_numbers_to_str)
-            .map(|val_match| val_match.unpack(state).into_py(py))
+            .map(|val_match| val_match.unpack(state).as_py_string(py, state.cache_str()).into_py(py))
     }
 
     fn get_name(&self) -> &str {
@@ -73,12 +73,12 @@ pub struct StrConstrainedValidator {
 impl_py_gc_traverse!(StrConstrainedValidator {});
 
 impl Validator for StrConstrainedValidator {
-    fn validate<'data>(
+    fn validate<'py>(
         &self,
-        py: Python<'data>,
-        input: &'data impl Input<'data>,
-        state: &mut ValidationState,
-    ) -> ValResult<'data, PyObject> {
+        py: Python<'py>,
+        input: &(impl Input<'py> + ?Sized),
+        state: &mut ValidationState<'_, 'py>,
+    ) -> ValResult<PyObject> {
         let either_str = input
             .validate_str(state.strict_or(self.strict), self.coerce_numbers_to_str)?
             .unpack(state);
@@ -129,14 +129,14 @@ impl Validator for StrConstrainedValidator {
         }
 
         let py_string = if self.to_lower {
-            PyString::new(py, &str.to_lowercase())
+            state.maybe_cached_str(py, &str.to_lowercase())
         } else if self.to_upper {
-            PyString::new(py, &str.to_uppercase())
+            state.maybe_cached_str(py, &str.to_uppercase())
         } else if self.strip_whitespace {
-            PyString::new(py, str)
+            state.maybe_cached_str(py, str)
         } else {
             // we haven't modified the string, return the original as it might be a PyString
-            either_str.as_py_string(py)
+            either_str.as_py_string(py, state.cache_str())
         };
         Ok(py_string.into_py(py))
     }
@@ -147,15 +147,23 @@ impl Validator for StrConstrainedValidator {
 }
 
 impl StrConstrainedValidator {
-    fn build(schema: &PyDict, config: Option<&PyDict>) -> PyResult<Self> {
+    fn build(schema: &Bound<'_, PyDict>, config: Option<&Bound<'_, PyDict>>) -> PyResult<Self> {
         let py = schema.py();
 
         let pattern = schema
             .get_as(intern!(py, "pattern"))?
             .map(|s| {
-                let regex_engine =
-                    schema_or_config(schema, config, intern!(py, "regex_engine"), intern!(py, "regex_engine"))?
-                        .unwrap_or(RegexEngine::RUST_REGEX);
+                let regex_engine = schema_or_config::<Bound<'_, PyString>>(
+                    schema,
+                    config,
+                    intern!(py, "regex_engine"),
+                    intern!(py, "regex_engine"),
+                )?;
+                let regex_engine = regex_engine
+                    .as_ref()
+                    .map(|s| s.to_str())
+                    .transpose()?
+                    .unwrap_or(RegexEngine::RUST_REGEX);
                 Pattern::compile(py, s, regex_engine)
             })
             .transpose()?;
@@ -176,10 +184,8 @@ impl StrConstrainedValidator {
         let to_upper: bool =
             schema_or_config(schema, config, intern!(py, "to_upper"), intern!(py, "str_to_upper"))?.unwrap_or(false);
 
-        let coerce_numbers_to_str = match config {
-            Some(c) => c.get_item("coerce_numbers_to_str")?.map_or(Ok(false), PyAny::is_true)?,
-            None => false,
-        };
+        let coerce_numbers_to_str: bool =
+            schema_or_config_same(schema, config, intern!(py, "coerce_numbers_to_str"))?.unwrap_or(false);
 
         Ok(Self {
             strict: is_strict(schema, config)?,
@@ -229,7 +235,7 @@ impl Pattern {
                 RegexEngine::RustRegex(Regex::new(&pattern).map_err(|e| py_schema_error_type!("{}", e))?)
             }
             RegexEngine::PYTHON_RE => {
-                let re_compile = py.import(intern!(py, "re"))?.getattr(intern!(py, "compile"))?;
+                let re_compile = py.import_bound(intern!(py, "re"))?.getattr(intern!(py, "compile"))?;
                 RegexEngine::PythonRe(re_compile.call1((&pattern,))?.into())
             }
             _ => return Err(py_schema_error_type!("Invalid regex engine: {}", engine)),
