@@ -1,5 +1,6 @@
 import re
 from decimal import Decimal
+from numbers import Number
 from typing import Any, Dict, Union
 
 import pytest
@@ -166,7 +167,8 @@ def test_str_constrained_config():
         v.validate_python('test long')
 
 
-def test_invalid_regex():
+@pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
+def test_invalid_regex(engine):
     # TODO uncomment and fix once #150 is done
     # with pytest.raises(SchemaError) as exc_info:
     #     SchemaValidator({'type': 'str', 'pattern': 123})
@@ -174,18 +176,25 @@ def test_invalid_regex():
     #     'Error building "str" validator:\n  TypeError: \'int\' object cannot be converted to \'PyString\''
     # )
     with pytest.raises(SchemaError) as exc_info:
-        SchemaValidator({'type': 'str', 'pattern': '(abc'})
-    assert exc_info.value.args[0] == (
-        'Error building "str" validator:\n'
-        '  SchemaError: regex parse error:\n'
-        '    (abc\n'
-        '    ^\n'
-        'error: unclosed group'
-    )
+        SchemaValidator(core_schema.str_schema(pattern='(abc', regex_engine=engine))
+
+    if engine is None or engine == 'rust-regex':
+        assert exc_info.value.args[0] == (
+            'Error building "str" validator:\n'
+            '  SchemaError: regex parse error:\n'
+            '    (abc\n'
+            '    ^\n'
+            'error: unclosed group'
+        )
+    elif engine == 'python-re':
+        assert exc_info.value.args[0] == (
+            'Error building "str" validator:\n  error: missing ), unterminated subpattern at position 0'
+        )
 
 
-def test_regex_error():
-    v = SchemaValidator({'type': 'str', 'pattern': '11'})
+@pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
+def test_regex_error(engine):
+    v = SchemaValidator(core_schema.str_schema(pattern='11', regex_engine=engine))
     with pytest.raises(ValidationError) as exc_info:
         v.validate_python('12')
     assert exc_info.value.errors(include_url=False) == [
@@ -201,7 +210,10 @@ def test_regex_error():
 
 def test_default_validator():
     v = SchemaValidator(core_schema.str_schema(strict=True, to_lower=False), {'str_strip_whitespace': False})
-    assert plain_repr(v) == 'SchemaValidator(title="str",validator=Str(StrValidator{strict:true}),definitions=[])'
+    assert (
+        plain_repr(v)
+        == 'SchemaValidator(title="str",validator=Str(StrValidator{strict:true,coerce_numbers_to_str:false}),definitions=[],cache_strings=True)'
+    )
 
 
 @pytest.fixture(scope='session', name='FruitEnum')
@@ -233,10 +245,25 @@ def test_lax_subclass(FruitEnum, kwargs):
     assert v.validate_python(b'foobar') == 'foobar'
     p = v.validate_python(FruitEnum.pear)
     assert p == 'pear'
-    assert type(p) is str  # noqa: E721
+    assert type(p) is str
     assert repr(p) == "'pear'"
 
 
+@pytest.mark.parametrize('kwargs', [{}, {'to_lower': True}], ids=repr)
+def test_lax_subclass_plain_enum(kwargs):
+    v = SchemaValidator(core_schema.str_schema(**kwargs))
+
+    from enum import Enum
+
+    class PlainEnum(Enum):
+        ONE = 'one'
+
+    p = v.validate_python(PlainEnum.ONE)
+    assert p == 'one'
+    assert type(p) is str
+    assert repr(p) == "'one'"
+
+
 def test_subclass_preserved() -> None:
     class StrSubclass(str):
         pass
@@ -253,3 +280,121 @@ def test_subclass_preserved() -> None:
 
     assert not isinstance(v.validate_python(StrSubclass('')), StrSubclass)
     assert not isinstance(v.validate_python(StrSubclass(''), strict=True), StrSubclass)
+
+
+def test_coerce_numbers_to_str_disabled_in_strict_mode() -> None:
+    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
+
+    v = SchemaValidator(core_schema.str_schema(strict=True), config)
+    with pytest.raises(ValidationError):
+        v.validate_python(42)
+    with pytest.raises(ValidationError):
+        v.validate_json('42')
+
+
+def test_coerce_numbers_to_str_raises_for_bool() -> None:
+    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
+
+    v = SchemaValidator(core_schema.str_schema(), config)
+    with pytest.raises(ValidationError):
+        v.validate_python(True)
+    with pytest.raises(ValidationError):
+        v.validate_json(False)
+
+
+@pytest.mark.parametrize(
+    ('number', 'expected_str'),
+    [
+        pytest.param(42, '42', id='42'),
+        pytest.param(42.0, '42.0', id='42.0'),
+        pytest.param(Decimal('42.0'), '42.0', id="Decimal('42.0')"),
+    ],
+)
+def test_coerce_numbers_to_str(number: Number, expected_str: str) -> None:
+    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
+
+    v = SchemaValidator(core_schema.str_schema(), config)
+    assert v.validate_python(number) == expected_str
+
+
+@pytest.mark.parametrize(
+    ('number', 'expected_str'),
+    [
+        pytest.param('42', '42', id='42'),
+        pytest.param('42.0', '42', id='42.0'),
+        pytest.param('42.13', '42.13', id='42.13'),
+    ],
+)
+def test_coerce_numbers_to_str_from_json(number: str, expected_str: str) -> None:
+    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
+
+    v = SchemaValidator(core_schema.str_schema(), config)
+    assert v.validate_json(number) == expected_str
+
+
+@pytest.mark.parametrize('mode', (None, 'schema', 'config'))
+def test_backtracking_regex_rust_unsupported(mode) -> None:
+    pattern = r'r(#*)".*?"\1'
+
+    with pytest.raises(SchemaError) as exc_info:
+        if mode is None:
+            # rust-regex is the default
+            SchemaValidator(core_schema.str_schema(pattern=pattern))
+        elif mode == 'schema':
+            SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='rust-regex'))
+        elif mode == 'config':
+            SchemaValidator(core_schema.str_schema(pattern=pattern), core_schema.CoreConfig(regex_engine='rust-regex'))
+
+    assert exc_info.value.args[0] == (
+        'Error building "str" validator:\n'
+        '  SchemaError: regex parse error:\n'
+        '    r(#*)".*?"\\1\n'
+        '              ^^\n'
+        'error: backreferences are not supported'
+    )
+
+
+@pytest.mark.parametrize('mode', ('schema', 'config'))
+def test_backtracking_regex_python(mode) -> None:
+    pattern = r'r(#*)".*?"\1'
+
+    if mode == 'schema':
+        v = SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='python-re'))
+    elif mode == 'config':
+        v = SchemaValidator(core_schema.str_schema(pattern=pattern), core_schema.CoreConfig(regex_engine='python-re'))
+    assert v.validate_python('r""') == 'r""'
+    assert v.validate_python('r#""#') == 'r#""#'
+    with pytest.raises(ValidationError):
+        # not a valid match for the pattern
+        v.validate_python('r#"#')
+
+
+@pytest.mark.parametrize('number', (42, 443, 10242))
+def test_coerce_numbers_to_str_schema(number: int):
+    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True))
+    assert v.validate_python(number) == str(number)
+    assert v.validate_json(str(number)) == str(number)
+
+
+@pytest.mark.parametrize('number', (42, 443, 10242))
+def test_coerce_numbers_to_str_schema_precedence(number: int):
+    config = core_schema.CoreConfig(coerce_numbers_to_str=False)
+    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True), config=config)
+    assert v.validate_python(number) == str(number)
+    assert v.validate_json(str(number)) == str(number)
+
+    config = core_schema.CoreConfig(coerce_numbers_to_str=True)
+    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=False), config=config)
+    with pytest.raises(ValidationError):
+        v.validate_python(number)
+    with pytest.raises(ValidationError):
+        v.validate_json(str(number))
+
+
+@pytest.mark.parametrize('number', (42, 443, 10242))
+def test_coerce_numbers_to_str_schema_with_strict_mode(number: int):
+    v = SchemaValidator(core_schema.str_schema(coerce_numbers_to_str=True, strict=True))
+    with pytest.raises(ValidationError):
+        v.validate_python(number)
+    with pytest.raises(ValidationError):
+        v.validate_json(str(number))
